"""Methods to add arguments based on class/method/function signatures."""

import dataclasses
import inspect
import re
from argparse import SUPPRESS, ArgumentParser
from contextlib import suppress
from typing import Any, Callable, List, Optional, Set, Tuple, Type, Union

from ._actions import _ActionConfigLoad
from ._common import (
    LoggerProperty,
    get_class_instantiator,
    get_generic_origin,
    get_unaliased_type,
    is_dataclass_like,
    is_subclass,
)
from ._namespace import Namespace
from ._optionals import get_doc_short_description, is_pydantic_model, pydantic_support
from ._parameter_resolvers import (
    ParamData,
    get_parameter_origins,
    get_signature_parameters,
)
from ._typehints import (
    ActionTypeHint,
    LazyInitBaseClass,
    callable_instances,
    get_subclasses_from_type,
    is_optional,
)
from ._util import NoneType, get_private_kwargs, iter_to_set_str
from .typing import register_pydantic_type

__all__ = [
    "compose_dataclasses",
    "SignatureArguments",
]


kinds = inspect._ParameterKind
inspect_empty = inspect._empty


class SignatureArguments(LoggerProperty):
    """Methods to add arguments based on signatures to an ArgumentParser instance."""

    def add_class_arguments(
        self,
        theclass: Type,
        nested_key: Optional[str] = None,
        as_group: bool = True,
        as_positional: bool = False,
        default: Optional[Union[dict, Namespace, LazyInitBaseClass]] = None,
        skip: Optional[Set[Union[str, int]]] = None,
        instantiate: bool = True,
        fail_untyped: bool = True,
        sub_configs: bool = False,
        **kwargs,
    ) -> List[str]:
        """Adds arguments from a class based on its type hints and docstrings.

        Note: Keyword arguments without at least one valid type are ignored.

        Args:
            theclass: Class from which to add arguments.
            nested_key: Key for nested namespace.
            as_group: Whether arguments should be added to a new argument group.
            as_positional: Whether to add required parameters as positional arguments.
            default: Default value used to override parameter defaults.
            skip: Names of parameters or number of positionals that should be skipped.
            instantiate: Whether the class group should be instantiated by :code:`instantiate_classes`.
            fail_untyped: Whether to raise exception if a required parameter does not have a type.
            sub_configs: Whether subclass type hints should be loadable from inner config file.

        Returns:
            The list of arguments added.

        Raises:
            ValueError: When not given a class.
            ValueError: When there are required parameters without at least one valid type.
        """
        if not inspect.isclass(get_generic_origin(get_unaliased_type(theclass))):
            raise ValueError(f"Expected 'theclass' parameter to be a class type, got: {theclass}")
        if not (
            isinstance(default, (NoneType, dict, Namespace))
            or (isinstance(default, LazyInitBaseClass) and isinstance(default, theclass))
        ):
            raise ValueError(
                f"Expected 'default' parameter to be a dict, Namespace or lazy instance of the class, got: {default}"
            )
        linked_targets = get_private_kwargs(kwargs, linked_targets=None)

        added_args = self._add_signature_arguments(
            theclass,
            None,
            nested_key,
            as_group,
            as_positional,
            skip,
            fail_untyped,
            sub_configs=sub_configs,
            instantiate=instantiate,
            linked_targets=linked_targets,
        )

        if default:
            skip = skip or set()
            prefix = nested_key + "." if nested_key else ""
            defaults = default
            if isinstance(default, LazyInitBaseClass):
                defaults = default.lazy_get_init_args().as_dict()
            elif isinstance(default, Namespace):
                defaults = default.as_dict()
            if defaults:
                defaults = {prefix + k: v for k, v in defaults.items() if k not in skip}
                self.set_defaults(**defaults)  # type: ignore[attr-defined]

        return added_args

    def add_method_arguments(
        self,
        theclass: Type,
        themethod: str,
        nested_key: Optional[str] = None,
        as_group: bool = True,
        as_positional: bool = False,
        skip: Optional[Set[Union[str, int]]] = None,
        fail_untyped: bool = True,
        sub_configs: bool = False,
    ) -> List[str]:
        """Adds arguments from a class based on its type hints and docstrings.

        Note: Keyword arguments without at least one valid type are ignored.

        Args:
            theclass: Class which includes the method.
            themethod: Name of the method for which to add arguments.
            nested_key: Key for nested namespace.
            as_group: Whether arguments should be added to a new argument group.
            as_positional: Whether to add required parameters as positional arguments.
            skip: Names of parameters or number of positionals that should be skipped.
            fail_untyped: Whether to raise exception if a required parameter does not have a type.
            sub_configs: Whether subclass type hints should be loadable from inner config file.

        Returns:
            The list of arguments added.

        Raises:
            ValueError: When not given a class or the name of a method of the class.
            ValueError: When there are required parameters without at least one valid type.
        """
        unaliased_type = get_unaliased_type(theclass)
        if not inspect.isclass(get_generic_origin(unaliased_type)):
            raise ValueError('Expected "theclass" argument to be a class object.')
        if not hasattr(unaliased_type, themethod) or not callable(getattr(unaliased_type, themethod)):
            raise ValueError('Expected "themethod" argument to be a callable member of the class.')

        return self._add_signature_arguments(
            theclass,
            themethod,
            nested_key,
            as_group,
            as_positional,
            skip,
            fail_untyped,
            sub_configs=sub_configs,
        )

    def add_function_arguments(
        self,
        function: Callable,
        nested_key: Optional[str] = None,
        as_group: bool = True,
        as_positional: bool = False,
        skip: Optional[Set[Union[str, int]]] = None,
        fail_untyped: bool = True,
        sub_configs: bool = False,
    ) -> List[str]:
        """Adds arguments from a function based on its type hints and docstrings.

        Note: Keyword arguments without at least one valid type are ignored.

        Args:
            function: Function from which to add arguments.
            nested_key: Key for nested namespace.
            as_group: Whether arguments should be added to a new argument group.
            as_positional: Whether to add required parameters as positional arguments.
            skip: Names of parameters or number of positionals that should be skipped.
            fail_untyped: Whether to raise exception if a required parameter does not have a type.
            sub_configs: Whether subclass type hints should be loadable from inner config file.

        Returns:
            The list of arguments added.

        Raises:
            ValueError: When not given a callable.
            ValueError: When there are required parameters without at least one valid type.
        """
        if not callable(function):
            raise ValueError('Expected "function" argument to be a callable object.')

        method_name = None
        if hasattr(function, "__class__") and callable_instances(function.__class__):
            function = function.__class__
            method_name = "__call__"

        return self._add_signature_arguments(
            function,
            method_name,
            nested_key,
            as_group,
            as_positional,
            skip,
            fail_untyped,
            sub_configs=sub_configs,
        )

    def _add_signature_arguments(
        self,
        function_or_class,
        method_name,
        nested_key: Optional[str],
        as_group: bool = True,
        as_positional: bool = False,
        skip: Optional[Set[Union[str, int]]] = None,
        fail_untyped: bool = True,
        sub_configs: bool = False,
        instantiate: bool = True,
        linked_targets: Optional[Set[str]] = None,
    ) -> List[str]:
        """Adds arguments from parameters of objects based on signatures and docstrings.

        Args:
            function_or_class: Object from which to add arguments.
            method_name: Class method from which to add arguments.
            nested_key: Key for nested namespace.
            as_group: Whether arguments should be added to a new argument group.
            as_positional: Whether to add required parameters as positional arguments.
            skip: Names of parameters or number of positionals that should be skipped.
            fail_untyped: Whether to raise exception if a required parameter does not have a type.
            sub_configs: Whether subclass type hints should be loadable from inner config file.
            instantiate: Whether the class group should be instantiated by :code:`instantiate_classes`.

        Returns:
            The list of arguments added.

        Raises:
            ValueError: When there are required parameters without at least one valid type.
        """
        params = get_signature_parameters(function_or_class, method_name, logger=self.logger)

        skip_positionals = [s for s in (skip or []) if isinstance(s, int)]
        if skip_positionals:
            if len(skip_positionals) > 1 or any(p <= 0 for p in skip_positionals):
                raise ValueError(f"Unexpected number of positionals to skip: {skip_positionals}")
            names = {p.name for p in params[: skip_positionals[0]]}
            params = params[skip_positionals[0] :]
            self.logger.debug(
                f"Skipping parameters {names} because {skip_positionals[0]} positionals requested to be skipped."
            )

        prefix = "--" + (nested_key + "." if nested_key else "")
        for param in params:
            if skip and param.name in skip:
                continue
            if prefix + param.name in self._option_string_actions:  # type: ignore[attr-defined]
                raise ValueError(
                    f"Unable to add parameter '{param.name}' from {function_or_class} because "
                    f"argument '{prefix + param.name}' already exists."
                )

        ## Create group if requested ##
        doc_group = get_doc_short_description(function_or_class, method_name, logger=self.logger)
        component = getattr(function_or_class, method_name) if method_name else function_or_class
        container = self._create_group_if_requested(
            component,
            nested_key,
            as_group,
            doc_group,
            config_load=len(params) > 0,
            instantiate=instantiate,
        )

        ## Add parameter arguments ##
        added_args: List[str] = []

        for param in params:
            self._add_signature_parameter(
                container,
                nested_key,
                param,
                added_args,
                skip={s for s in (skip or []) if isinstance(s, str)},
                fail_untyped=fail_untyped,
                sub_configs=sub_configs,
                linked_targets=linked_targets,
                as_positional=as_positional,
            )

        if hasattr(function_or_class, "__scriptconfig__"):
            # Integrate with scriptconfig style classes.
            # When a function/class has a __scriptconfig__ object that means it
            # should accept any of the options defined in the config as keyword
            # arguments. We can utilize additional metadata stored in the
            # scriptconfig object to enrich our CLI.
            config_cls = function_or_class.__scriptconfig__
            self._add_scriptconfig_arguments(
                config_cls=config_cls,
                added_args=added_args,
                function_or_class=function_or_class,
                component=component,
                container=container,
                nested_key=nested_key,
                fail_untyped=fail_untyped,
                as_positional=as_positional,
                sub_configs=sub_configs,
                skip=skip,
                linked_targets=linked_targets,
            )

        return added_args

    def _add_scriptconfig_arguments(
        self,
        config_cls,
        added_args,
        function_or_class,
        component,
        container,
        nested_key,
        fail_untyped: bool = True,
        as_positional: bool = False,
        sub_configs: bool = False,
        skip: Optional[Set[Union[str, int]]] = None,
        linked_targets: Optional[Set[str]] = None,
    ):

        for key, value in config_cls.__default__.items():
            from scriptconfig import value as value_mod

            if not isinstance(value, value_mod.Value):
                # hack
                value = value_mod.Value(value)
            type_ = value.parsekw["type"]
            if type_ is None or not isinstance(type_, type):
                annotation = inspect._empty
            else:
                annotation = type_
            param = ParamData(
                name=key,
                annotation=annotation,
                kind=inspect.Parameter.KEYWORD_ONLY,
                default=value.value,
                doc=value.parsekw["help"],
                component=component,
                parent=function_or_class,
                short_aliases=value.short_alias,
                long_aliases=value.alias,
            )
            self._add_signature_parameter(
                container,
                nested_key,
                param,
                added_args,
                skip={s for s in (skip or []) if isinstance(s, str)},
                fail_untyped=fail_untyped,
                sub_configs=sub_configs,
                linked_targets=linked_targets,
                as_positional=as_positional,
            )

    def _add_signature_parameter(
        self,
        container,
        nested_key: Optional[str],
        param,
        added_args: List[str],
        skip: Optional[Set[str]] = None,
        fail_untyped: bool = True,
        as_positional: bool = False,
        sub_configs: bool = False,
        instantiate: bool = True,
        linked_targets: Optional[Set[str]] = None,
        default: Any = inspect_empty,
        **kwargs,
    ):
        name = param.name
        kind = param.kind
        annotation = param.annotation
        if default == inspect_empty:
            default = param.default
            if default == inspect_empty and is_optional(annotation):
                default = None
        is_required = default == inspect_empty
        src = get_parameter_origins(param.component, param.parent)
        skip_message = f'Skipping parameter "{name}" from "{src}" because of: '
        if not fail_untyped and annotation == inspect_empty:
            annotation = Any
            default = None if is_required else default
            is_required = False
        if is_required and linked_targets is not None and name in linked_targets:
            default = None
            is_required = False
        if (
            kind in {kinds.VAR_POSITIONAL, kinds.VAR_KEYWORD}
            or (not is_required and name[0] == "_")
            or (annotation == inspect_empty and not is_required and default is None)
        ):
            return
        elif skip and name in skip:
            self.logger.debug(skip_message + "Parameter requested to be skipped.")
            return
        if is_factory_class(default):
            default = param.parent.__dataclass_fields__[name].default_factory()
        if annotation == inspect_empty and not is_required:
            annotation = Union[type(default), Any]
        if "help" not in kwargs:
            kwargs["help"] = param.doc
        if not is_required:
            kwargs["default"] = default
            if default is None and not is_optional(annotation, object):
                annotation = Optional[annotation]
        elif not as_positional:
            kwargs["required"] = True
        is_subclass_typehint = False
        is_dataclass_like_typehint = is_dataclass_like(annotation)
        # dest = (nested_key + "." if nested_key else "") + name
        # args = [dest if is_required and as_positional else "--" + dest]
        args, dest = param._resolve_args_and_dest(is_required, as_positional, nested_key)
        if param.origin:
            parser = container
            if not isinstance(container, ArgumentParser):
                parser = getattr(container, "parser")
            group_name = "; ".join(str(o) for o in param.origin)
            if group_name in parser.groups:
                container = parser.groups[group_name]
            else:
                container = parser.add_argument_group(
                    f"Conditional arguments [origins: {group_name}]",
                    name=group_name,
                )
        if (
            annotation in {str, int, float, bool}
            or is_subclass(annotation, (str, int, float))
            or is_dataclass_like_typehint
        ):
            kwargs["type"] = annotation
            register_pydantic_type(annotation)
        elif annotation != inspect_empty:
            try:
                is_subclass_typehint = ActionTypeHint.is_subclass_typehint(annotation, all_subtypes=False)
                kwargs["type"] = annotation
                sub_add_kwargs: dict = {"fail_untyped": fail_untyped, "sub_configs": sub_configs}
                if is_subclass_typehint:
                    prefix = f"{name}.init_args."
                    subclass_skip = {s[len(prefix) :] for s in skip or [] if s.startswith(prefix)}
                    sub_add_kwargs["skip"] = subclass_skip
                else:
                    register_pydantic_type(annotation)
                enable_path = sub_configs and (
                    is_subclass_typehint or ActionTypeHint.is_return_subclass_typehint(annotation)
                )
                args = ActionTypeHint.prepare_add_argument(
                    args=args,
                    kwargs=kwargs,
                    enable_path=enable_path,
                    container=container,
                    logger=self.logger,
                    sub_add_kwargs=sub_add_kwargs,
                )
            except ValueError as ex:
                self.logger.debug(skip_message + str(ex))
        if "type" in kwargs or "action" in kwargs:
            sub_add_kwargs = {
                "fail_untyped": fail_untyped,
                "sub_configs": sub_configs,
                "instantiate": instantiate,
            }
            if is_dataclass_like_typehint:
                kwargs.update(sub_add_kwargs)
            with ActionTypeHint.allow_default_instance_context():
                action = container.add_argument(*args, **kwargs)
            action.sub_add_kwargs = sub_add_kwargs
            if is_subclass_typehint and len(subclass_skip) > 0:
                action.sub_add_kwargs["skip"] = subclass_skip
            added_args.append(dest)
        elif is_required and fail_untyped:
            raise ValueError(
                "With fail_untyped=True, all mandatory parameters must have a supported"
                f" type. Parameter '{name}' from '{src}' does not specify a type."
            )

    def add_dataclass_arguments(
        self,
        theclass: Type,
        nested_key: str,
        default: Optional[Union[Type, dict]] = None,
        as_group: bool = True,
        fail_untyped: bool = True,
        **kwargs,
    ) -> List[str]:
        """Adds arguments from a dataclass based on its field types and docstrings.

        Args:
            theclass: Class from which to add arguments.
            nested_key: Key for nested namespace.
            default: Value for defaults. Must be instance of or kwargs for theclass.
            as_group: Whether arguments should be added to a new argument group.
            fail_untyped: Whether to raise exception if a required parameter does not have a type.

        Returns:
            The list of arguments added.

        Raises:
            ValueError: When not given a dataclass.
            ValueError: When default is not instance of or kwargs for theclass.
        """
        if not is_dataclass_like(theclass):
            raise ValueError(f'Expected "theclass" argument to be a dataclass-like, given {theclass}')

        doc_group = get_doc_short_description(theclass, logger=self.logger)
        for key in ["help", "title"]:
            if key in kwargs and kwargs[key] is not None:
                doc_group = kwargs.pop(key)
        group = self._create_group_if_requested(theclass, nested_key, as_group, doc_group, config_load_type=theclass)

        defaults = {}
        if default is not None:
            if isinstance(default, dict):
                with suppress(TypeError):
                    default = theclass(**default)
            if not isinstance(default, get_unaliased_type(theclass)):
                raise ValueError(
                    f'Expected "default" argument to be an instance of "{theclass.__name__}" '
                    f"or its kwargs dict, given {default}"
                )
            defaults = dataclass_to_dict(default)

        added_args: List[str] = []
        param_kwargs = {k: v for k, v in kwargs.items() if k == "sub_configs"}
        for param in get_signature_parameters(theclass, None, logger=self.logger):
            self._add_signature_parameter(
                group,
                nested_key,
                param,
                added_args,
                fail_untyped=fail_untyped,
                default=defaults.get(param.name, inspect_empty),
                **param_kwargs,
            )

        return added_args

    def add_subclass_arguments(
        self,
        baseclass: Union[Type, Tuple[Type, ...]],
        nested_key: str,
        as_group: bool = True,
        skip: Optional[Set[str]] = None,
        instantiate: bool = True,
        required: bool = False,
        metavar: str = "CONFIG | CLASS_PATH_OR_NAME | .INIT_ARG_NAME VALUE",
        help: str = (
            'One or more arguments specifying "class_path" and "init_args" for any subclass of %(baseclass_name)s.'
        ),
        **kwargs,
    ):
        """Adds arguments to allow specifying any subclass of the given base class.

        This adds an argument that requires a dictionary with a "class_path"
        entry which must be a import dot notation expression. Optionally any
        init arguments for the class can be given in the "init_args" entry.
        Since subclasses can have different init arguments, the help does not
        show the details of the arguments of the base class. Instead a help
        argument is added that will print the details for a given class path.

        Args:
            baseclass: Base class or classes to use to check subclasses.
            nested_key: Key for nested namespace.
            as_group: Whether arguments should be added to a new argument group.
            skip: Names of parameters that should be skipped.
            required: Whether the argument group is required.
            metavar: Variable string to show in the argument's help.
            help: Description of argument to show in the help.
            **kwargs: Additional parameters like in add_class_arguments.

        Raises:
            ValueError: When given an invalid base class.
        """
        if is_dataclass_like(baseclass):
            raise ValueError("Not allowed for dataclass-like classes.")
        if type(baseclass) is not tuple:
            baseclass = (baseclass,)  # type: ignore[assignment]
        if not baseclass or not all(ActionTypeHint.is_subclass_typehint(c, also_lists=True) for c in baseclass):
            raise ValueError(f"Expected 'baseclass' to be a subclass type or a tuple of subclass types: {baseclass}")

        doc_group = None
        if len(baseclass) == 1:  # type: ignore[arg-type]
            doc_group = get_doc_short_description(baseclass[0], logger=self.logger)
        group = self._create_group_if_requested(
            baseclass,
            nested_key,
            as_group,
            doc_group,
            config_load=False,
            required=required,
            instantiate=False,
        )

        added_args: List[str] = []
        if skip is not None:
            skip = {f"{nested_key}.init_args." + s for s in skip}
        param = ParamData(name=nested_key, annotation=Union[baseclass], component=baseclass)
        str_baseclass = iter_to_set_str(get_subclasses_from_type(param.annotation))
        kwargs.update(
            {
                "metavar": metavar,
                "help": (help % {"baseclass_name": str_baseclass}),
            }
        )
        if "default" not in kwargs:
            kwargs["default"] = SUPPRESS
        self._add_signature_parameter(
            group, None, param, added_args, skip, sub_configs=True, instantiate=instantiate, **kwargs
        )

    def _create_group_if_requested(
        self,
        obj,
        nested_key,
        as_group,
        doc_group,
        config_load=True,
        config_load_type=None,
        required=False,
        instantiate=True,
    ):
        if required:
            if nested_key is None:
                raise ValueError("A nested_key is mandatory to make required.")
            self.required_args.add(nested_key)

        group = self
        if as_group:
            if doc_group is None:
                if isinstance(obj, tuple) and len(obj) == 1:
                    doc_group = str(obj[0])
                else:
                    doc_group = str(obj)
            name = obj.__name__ if nested_key is None else nested_key
            group = self.add_argument_group(strip_title(doc_group), name=name)
            if config_load and nested_key is not None:
                group.add_argument("--" + nested_key, action=_ActionConfigLoad(basetype=config_load_type))
            if inspect.isclass(obj) and nested_key is not None and instantiate:
                group.dest = nested_key
                group.group_class = obj
                group.instantiate_class = group_instantiate_class
        return group


def group_instantiate_class(group, cfg):
    try:
        value, parent, key = cfg.get_value_and_parent(group.dest)
    except KeyError:
        value = {}
        parent = cfg
        key = group.dest
    instantiator_fn = get_class_instantiator()
    parent[key] = instantiator_fn(group.group_class, **value)


def strip_title(value):
    if value is not None:
        value = re.sub(r"\.$", "", value.strip())
    return value


def is_factory_class(value):
    return value.__class__ == dataclasses._HAS_DEFAULT_FACTORY_CLASS


def dataclass_to_dict(value) -> dict:
    if pydantic_support:
        pydantic_model = is_pydantic_model(type(value))
        if pydantic_model:
            return value.dict() if pydantic_model == 1 else value.model_dump()
    if isinstance(value, LazyInitBaseClass):
        return value.lazy_get_init_data().as_dict()
    return dataclasses.asdict(value)


def compose_dataclasses(*args):
    """Returns a dataclass inheriting all given dataclasses and properly handling __post_init__."""

    @dataclasses.dataclass
    class ComposedDataclass(*args):
        def __post_init__(self):
            for arg in args:
                if hasattr(arg, "__post_init__"):
                    arg.__post_init__(self)

    return ComposedDataclass



def patch_kwutil_toplevel():
    """
    For kwutil versions <= 0.2.6 which don't expose common methods at the top
    level.

    Ignore:
        from geowatch.monkey import monkey_kwutil
        monkey_kwutil.patch_kwutil_toplevel()
    """
    import kwutil
    kwutil.datetime = kwutil.util_time.datetime
    kwutil.timedelta = kwutil.util_time.timedelta
    kwutil.Yaml = kwutil.util_yaml.Yaml



import torch


def coerce_criterion(loss_code, weights, ohem_ratio=None, focal_gamma=2.0,
                     spatial_dims='legacy'):
    """
    Helps build a loss function and returns information about the shapes needed
    by the specific loss. Augments the criterion with extra information about
    what it expects.

    Args:
        loss_code (str): The code that corresponds to loss function call.
            One of ['cce', 'focal', 'dicefocal'].
        weights (torch.Tensor): Per class weights.
            Note: Only used for 'cce' and 'focal' losses.
        ohem_ratio (float): Ratio of hard examples to sample to compute loss.
            Note: Only applies to focal losses.
        focal_gamma (float): Focal loss gamma parameter.

        spatial_dims (str):
            A code indicating which spatial dimension we are expecting in this
            loss. The "legacy" maintains backwards compat with the multimodal
            transformer. For spacetime segmentation this should usually be
            't h w'. For nonlocal it should be ''.

    Raises:
        KeyError: if loss_code is not recognized.

    Returns:
        torch.nn.modules.loss._Loss: The loss function.

        The loss criterion will contain variables:

            target_encoding: which is either index or onehot
            logit_shape: the expected shape of the predicted logits.
            target_shape: the expected shape of the truth targets.
    """
    # import monai
    if loss_code == 'cce':
        criterion = torch.nn.CrossEntropyLoss(
            weight=weights, reduction='none')
        criterion.target_encoding = 'index'
        if spatial_dims == 'legacy':
            spatial_dims = 't h w'
        criterion.logit_shape = f'(b {spatial_dims}) c'
        criterion.target_shape = f'(b {spatial_dims})'

    elif loss_code == 'bce':
        criterion = torch.nn.BCELoss(
            weight=weights, reduction='none')
        criterion.target_encoding = 'onehot'
        if spatial_dims == 'legacy':
            spatial_dims = 't h w'
        criterion.logit_shape = '(b {spatial_dims}) c'
        criterion.target_shape = f'(b {spatial_dims}) c'

    elif loss_code == 'focal_multiclass':
        from geowatch.utils.ext_monai import FocalLoss
        # from monai.losses import FocalLoss
        criterion = FocalLoss(
            reduction='none',
            to_onehot_y=True,
            weight=weights,
            ohem_ratio=ohem_ratio,
            gamma=focal_gamma)

        criterion.target_encoding = 'index'
        if spatial_dims == 'legacy':
            spatial_dims = 't h w'
        criterion.logit_shape = f'(b {spatial_dims}) c'
        criterion.target_shape = f'(b {spatial_dims})'

    elif loss_code == 'focal':
        from geowatch.utils.ext_monai import FocalLoss
        # from monai.losses import FocalLoss
        criterion = FocalLoss(
            reduction='none',
            to_onehot_y=False,
            weight=weights,
            ohem_ratio=ohem_ratio,
            gamma=focal_gamma)
        criterion.target_encoding = 'onehot'
        if spatial_dims == 'legacy':
            spatial_dims = 'h w t'
        criterion.logit_shape = f'b c {spatial_dims}'
        criterion.target_shape = f'b c {spatial_dims}'

    elif loss_code == 'dicefocal':
        from geowatch.utils.ext_monai import DiceFocalLoss
        # from monai.losses import DiceFocalLoss
        criterion = DiceFocalLoss(
            sigmoid=True,
            to_onehot_y=False,
            focal_weight=weights,
            reduction='none',
            ohem_ratio_focal=ohem_ratio,
            gamma=focal_gamma)
        criterion.target_encoding = 'onehot'
        if spatial_dims == 'legacy':
            spatial_dims = 'h w t'
        criterion.logit_shape = f'b c {spatial_dims}'
        criterion.target_shape = f'b c {spatial_dims}'
    else:
        # self.class_criterion = nn.CrossEntropyLoss()
        # self.class_criterion = nn.BCEWithLogitsLoss()
        raise NotImplementedError(loss_code)

    criterion.in_channels = len(weights)
    return criterion



#!/usr/bin/env python3
# PYTHON_ARGCOMPLETE_OK
import ubelt as ub

__devnotes__ = """
# We may want to delay actual imports, gdal import time can be excessive

python -X importtime -c "import geowatch"
GEOWATCH_HACK_IMPORT_ORDER="" python  -X importtime -m geowatch.cli find_dvc
"""


def main(cmdline=True, **kw):
    """
    The geowatch command line interface
    """
    modnames = [
        'watch_coco_stats',
        'geojson_site_stats',
        'validate_annotation_schemas',
        'torch_model_stats',
        'coco_spectra',
        'draw_region',

        # 'geotiffs_to_kwcoco'  # TODO: cleanup and add

        'coco_visualize_videos',

        'find_dvc',

        'coco_add_watch_fields',
        'coco_align',
        'coco_clean_geotiffs',
        'coco_average_features',
        'coco_time_combine',
        'coco_add_ignore_buffer',

        'reproject_annotations',

        'run_tracker',
        'run_metrics_framework',

        'crop_sites_to_regions',
        'coco_remove_bad_images',
        'coco_average_features',

        # Due to LightningCLI fit does not play nicely here.
        'fit',
        # Predict also has issues because of its heavy imports.
        'predict',

        # 'mlops_cli',
        'special.finish_install',
    ]

    cmd_alias = {
        'geowatch.cli.torch_model_stats': ['model_stats', 'model_info'],
        # 'geowatch.cli.geojson_site_stats': ['site_stats', 'geojson_stats', 'geomodel_stats'],
        'geowatch.cli.watch_coco_stats': ['stats'],
        'geowatch.cli.coco_visualize_videos': ['visualize'],
        'geowatch.cli.coco_align': ['align', 'coco_align_geotiffs'],
        'geowatch.cli.reproject_annotations': ['reproject', 'project'],
        'geowatch.cli.coco_add_watch_fields': ['add_fields'],
        'geowatch.cli.coco_spectra': ['spectra', 'intensity_histograms'],
        'geowatch.cli.run_metrics_framework': ['iarpa_eval'],
        'geowatch.cli.coco_clean_geotiffs': ['clean_geotiffs'],
        'geowatch.cli.run_tracker': ['run_tracker'],
        'geowatch.cli.find_dvc': ['dvc', 'dvcdir'],
        'geowatch.cli.coco_average_features': ['average_features', 'ensemble'],
        'geowatch.cli.coco_time_combine': ['time_combine'],
        'geowatch.cli.coco_add_ignore_buffer': ['ignore_buffer'],
        'geowatch.cli.crop_sites_to_regions': ['crop_sitemodels'],
        'geowatch.cli.coco_remove_bad_images': ['remove_bad_images'],
        'geowatch.cli.mlops_cli': ['mlops'],
        'geowatch.cli.special.finish_install': [],
        'geowatch.mlops.aggregate': ['mlops.aggregate'],
        'geowatch.mlops.schedule_evaluation': ['mlops.schedule'],
    }

    module_lut = {}
    for name in modnames:
        mod = ub.import_module_from_name('geowatch.cli.{}'.format(name))
        module_lut[name] = mod

    module_lut['schedule'] = ub.import_module_from_name('geowatch.mlops.schedule_evaluation')
    module_lut['manager'] = ub.import_module_from_name('geowatch.mlops.manager')
    module_lut['aggregate'] = ub.import_module_from_name('geowatch.mlops.aggregate')
    module_lut['repackage'] = ub.import_module_from_name('geowatch.mlops.repackager')
    # module_lut['fit'] = ub.import_module_from_name('geowatch.tasks.fusion.fit_lightning')
    # module_lut['predict'] = ub.import_module_from_name('geowatch.tasks.fusion.predict')

    # Create a list of all submodules with CLI interfaces
    cli_modules = list(module_lut.values())

    import os
    from scriptconfig.modal import ModalCLI
    import geowatch
    GEOWATCH_LOOSE_CLI = os.environ.get('WATCH_LOOSE_CLI', '') or os.environ.get('GEOWATCH_LOOSE_CLI', '')

    # https://emojiterra.com/time/
    # Not sure how to make this consistent on different terminals
    fun_header = ub.codeblock(
        '''
        🛰️⌚🛰️           🧠       🛰️⌚🛰️
        🌍🌎🌏        👁️   👁️      🌏🌎🌍
        ''')

    boring_description =  ub.codeblock(
        f'''
        🌐🌐🌐 The GeoWATCH CLI 🌐🌐🌐

        An open source research and production environment for image and video
        segmentation and detection with geospatial awareness.

        Developed by [link=https://www.kitware.com/]Kitware[/link]. Funded by the [link=https://www.iarpa.gov/research-programs/smart]IARPA SMART[/link] challenge.

        Version: {geowatch.__version__}
        ''')

    FUN = os.getenv('FUN', '') and not os.getenv('NOFUN', '')
    if FUN:
        description = fun_header + '\n' + boring_description
    else:
        description = boring_description

    modal = ModalCLI(description)

    # scriptconfig bug made this not work...
    # modal.__class__.version = geowatch.__version__

    def get_version(self):
        import geowatch
        return geowatch.__version__
    modal.__class__.version = property(get_version)

    for cli_module in cli_modules:

        cli_config = None

        if hasattr(cli_module, '__cli__'):
            # New way
            cli_config = cli_module.__cli__
        else:
            if hasattr(cli_module, 'modal'):
                continue
            else:
                raise NotImplementedError(f'modules must define the __cli__ attribute to be registered. Failed on {cli_module}')

        if not hasattr(cli_config, 'main'):
            if hasattr(cli_module, 'main'):
                main_func = cli_module.main
                # Hack the main function into the config
                cli_config.main = main_func
            else:
                raise AssertionError(f'No main function for {cli_module}')

        # Update configs to have aliases / commands attributes
        cli_modname = cli_module.__name__
        cli_rel_modname = cli_modname.split('.')[-1]

        cmdname_aliases = ub.oset()
        alias = getattr(cli_config, '__alias__', [])
        if isinstance(alias, str):
            alias = [alias]
        command = getattr(cli_module, '__command__', None)
        command = getattr(cli_config, '__command__', command)
        if command is not None:
            cmdname_aliases.add(command)
        cmdname_aliases.update(alias)
        cmdname_aliases.update(cmd_alias.get(cli_modname, []) )
        cmdname_aliases.add(cli_rel_modname)
        # parserkw = {}
        primary_cmdname = cmdname_aliases[0]
        secondary_cmdnames = cmdname_aliases[1:]
        if not isinstance(primary_cmdname, str):
            raise AssertionError(primary_cmdname)
        cli_config.__command__ = primary_cmdname
        cli_config.__alias__ = list(secondary_cmdnames)
        modal.register(cli_config)

    ret = modal.main(strict=not GEOWATCH_LOOSE_CLI)
    return ret


if __name__ == '__main__':
    """
    CommandLine:
        python -m geowatch.cli --help
        python -m geowatch.cli.watch_coco_stats
    """
    main()
    # sys.exit(main())



#!/usr/bin/env python3
import sys
import csv
import random
import glob

import numpy as np
import torch
from torch.utils.data import Dataset
from torchvision import transforms

from PIL import Image
from tqdm import tqdm

import json
import geojson
import rasterio
import rasterio.mask
from rasterio.windows import Window
from pyproj import CRS, Transformer

# from skimage.transform import resize

from matplotlib import pyplot as plt

from lydorn_utils import polygon_utils


class BuildingDataset(Dataset):

    def __init__(self, tier=1, show_mode=False, augment=False, small_subset=False, crop_size=1024, resize_size=224,
                 window_random_shift=2048, data_dir="./", baseline_mode=False, transform=None, val=False, val_split=0.1, split_seed=42, sampling_mode="polygons"):
        super().__init__()

        self.crop_size = crop_size
        self.resize_size = resize_size

        self.data_dir = data_dir

        random.seed(42)

        # Load TIF and geojson file names
        # from train_metadata.csv
        self.img_ids_to_tif = dict()
        self.img_ids_to_geojson = dict()
        with open(data_dir + "/train_metadata.csv", 'r') as f:
            csvreader = csv.reader(f)
            header = next(csvreader)  # NOQA
            imgs = []
            for row in csvreader:
                if int(row[3]) <= tier:
                    imgs.append(row)
            random.shuffle(imgs)
            n_train = int(len(imgs) * (1.0 - val_split))
            if val:
                imgs = imgs[n_train:]
            else:
                imgs = imgs[:n_train]
            for row in imgs:
                imgid = row[0].split("/")[2]
                self.img_ids_to_tif[imgid] = row[0]
                self.img_ids_to_geojson[imgid] = row[1]

        # Load all geojson files
        print("Loading geojson files")

        self.geojson_data = dict()
        self.all_polygons = []
        self.feat_to_img_id = []
        for ids, geojson_f in tqdm(self.img_ids_to_geojson.items()):
            self.geojson_data[ids] = self._load_geojsonfile(data_dir + "/" + geojson_f)
            self.all_polygons += self.geojson_data[ids]["features"]
            for f in self.geojson_data[ids]["features"]:
                self.feat_to_img_id.append(ids)

        print("Number of training polygons:", len(self.all_polygons))

        # Open all tif files
        # and get corresponding transformers
        print("Opening rasters")
        #self.rasters = dict()
        img_to_transformer_dict = dict()
        for ids, img_f in tqdm(self.img_ids_to_tif.items()):
            with rasterio.open(data_dir + "/" + img_f) as raster:
                img_to_transformer_dict[ids] = Transformer.from_crs(CRS.from_proj4("+proj=latlon"), raster.crs)

        print("Reading means and std")
        self.stats = dict()
        for ids, tifs in self.img_ids_to_tif.items():
            with open(data_dir + "/" + ".".join([tifs.split(".")[0], "tif.stats.json"]), "r") as statfile:
                self.stats[ids] = json.load(statfile)

        print("Reprojecting polygons")

        self.img_id_to_polys = dict()
        for k, v in self.img_ids_to_tif.items():
            self.img_id_to_polys[k] = []

        for i in tqdm(range(len(self.all_polygons))):
            self.convert_poly(i, img_to_transformer_dict)
            self.img_id_to_polys[self.feat_to_img_id[i]].append(self.all_polygons[i])

        if small_subset:
            self.all_polygons = self.all_polygons[:200]

        self.show_mode = show_mode
        self.baseline_mode = baseline_mode

        self.aug_transforms = transforms.Compose([transforms.RandomVerticalFlip(),
                                                  transforms.RandomHorizontalFlip(),
                                                  transforms.RandomAffine((180), (0.2, 0.2), (0.5, 2)),
                                                  transforms.ColorJitter(0.2, 0.2, 0.5, 0.2),
                                                  transforms.Resize((resize_size, resize_size)),
                                                  transforms.ToTensor(),
                                                  # transforms.RandomErasing(),
                                                  transforms.ToPILImage()])

        self.noaug_transforms = transforms.Compose([transforms.Resize((resize_size, resize_size))])

        self.mask_transforms = transforms.Compose([transforms.Resize((resize_size, resize_size)),
                                                   transforms.ToTensor()])

        self.img_transforms = transforms.Compose([transforms.ToTensor(),
                                                  transforms.Normalize(mean=[0.485, 0.456, 0.406],
                                                                       std=[0.229, 0.224, 0.225])])

        self.wow_transforms = transform

        self.augment = augment

        self.window_random_shift = window_random_shift

        self.sampling_mode = sampling_mode

    def __len__(self):
        if self.sampling_mode == "polygons":
            return len(self.all_polygons)
        else:
            return 1000

    def __getitem__(self, i):
        img_id = self.feat_to_img_id[i]

        raster = rasterio.open(self.data_dir + "/" + self.img_ids_to_tif[img_id])

        if self.sampling_mode == "random":
            window = self._get_random_window()
        else:
            window = self._get_window(raster, self.all_polygons[i])

        b1, b2, b3, b4 = raster.read(window=window)
        out_image = np.stack([b1, b2, b3])
        out_image = np.swapaxes(out_image, 0, 2)[:, :, 0:3]

        try:
            img = Image.fromarray(out_image, 'RGB')
        except Exception:
            print(window)
            print(raster.width, raster.height)

        # create rasterized edges
        polygons = []
        for feat in self.img_id_to_polys[img_id]:
            try:
                # polywindow = rasterio.features.geometry_window(raster, [feat])
                polywindow = self._get_window(raster, feat, padded=False)

                if rasterio.windows.intersect(polywindow, window):
                    if feat["geometry"]["type"] == "MultiPolygon":
                        local_poly = []
                        for p in feat["geomtry"]["coordinates"]:
                            local_poly += self._polygon_window_pixel_coords(raster, window, p[0])
                    elif feat["geometry"]["type"] == "Polygon":
                        local_poly = self._polygon_window_pixel_coords(raster, window,
                                                                       feat["geometry"]["coordinates"][0])
                    ratio = self.resize_size / self.crop_size
                    scaled_poly = []
                    for point in local_poly:
                        x, y = point
                        scaled_poly.append((x * ratio, y * ratio))
                    polygons.append(scaled_poly)
            except Exception:
                pass

        masks = polygon_utils.draw_polygons(polygons, (self.resize_size, self.resize_size), line_width=2, antialiasing=True)
        masks = np.moveaxis(np.array(masks), 2, 0)

        angles = polygon_utils.init_angle_field(polygons, (self.resize_size, self.resize_size), line_width=4)
        angles = np.expand_dims(angles, 0)

        if self.augment:
            trans = self.aug_transforms
        else:
            trans = self.noaug_transforms

        mean = [self.stats[img_id]["stats"]["mean_r"], self.stats[img_id]["stats"]["mean_g"], self.stats[img_id]["stats"]["mean_b"]]
        std = [self.stats[img_id]["stats"]["std_r"], self.stats[img_id]["stats"]["std_g"], self.stats[img_id]["stats"]["std_b"]]

        out_img = np.array(trans(img))
        out_img = torch.Tensor(np.moveaxis(out_img, -1, 0))

        if self.show_mode:
            return {"image": trans(img), "gt_polygons_image": masks, "gt_crossfield_angle": angles, "image_mean": mean,
                    "image_std": std}
        elif self.baseline_mode:
            masks = np.moveaxis(np.array(masks), 0, 2)
            masks = masks / 255
            return {"image": self.img_transforms(trans(img)), "gt_polygons_image": masks}
        else:
            return {"image": out_img,
                    "gt_polygons_image": torch.Tensor(masks),
                    "gt_crossfield_angle": torch.Tensor(angles),
                    "image_mean": torch.Tensor(mean),
                    "image_std": torch.Tensor(std)}

    def _get_window(self, raster, feat, padded=True):

        poly = feat["geometry"]["coordinates"]  # NOQA

        box = []

        if feat["geometry"]["type"] == "MultiPolygon":
            local_poly = []
            for p in feat["geometry"]["coordinates"]:
                local_poly += self._polygon_pixel_coords(raster, p[0])
        elif feat["geometry"]["type"] == "Polygon":
            local_poly = self._polygon_pixel_coords(raster, feat["geometry"]["coordinates"][0])

        for i in (0, 1):
            res = sorted(local_poly, key=lambda x: x[i])
            box += [res[0][i], res[-1][i]]

        if padded:
            toplx, toply = (box[0] + box[1]) / 2 - self.crop_size / 2, (box[2] + box[3]) / 2 - self.crop_size / 2

            while True:
                shift_y = random.randint(-self.window_random_shift, self.window_random_shift)
                shift_x = random.randint(-self.window_random_shift, self.window_random_shift)
                win = Window(toplx + shift_x, toply + shift_y, self.crop_size, self.crop_size)

                # check that the window is in the image
                if win.col_off + win.width < raster.width and win.col_off > 0:
                    if win.row_off + win.height < raster.height and win.row_off > 0:
                        break
            return win
        else:
            return Window(box[0], box[1], box[2] - box[0], box[3] - box[1])

    def _get_random_window(self):

        rastern = random.randint(0, len(self.img_ids_to_tif) - 1)

        img_id = list(self.img_ids_to_tif)[rastern]

        tiff = self.img_ids_to_tif[img_id]

        raster = rasterio.open(self.data_dir + "/" + tiff)

        while True:
            x = random.randint(0, raster.width - 1)
            y = random.randint(0, raster.height - 1)

            sample = [val for val in raster.sample([raster.transform * (x, y)])][0]
            if sample[3] == 255:
                win = Window(x - self.crop_size / 2, y - self.crop_size / 2, self.crop_size, self.crop_size)
                if win.col_off + win.width < raster.width and win.col_off > 0:
                    if win.row_off + win.height < raster.height and win.row_off > 0:
                        return win, img_id

    def _polygon_pixel_coords(self, raster, poly):
        # converts to raster pixel coords from utm
        local_poly = []
        for point in poly:
            local_poly.append(~raster.transform * (point[0], point[1]))

        return local_poly

    def _polygon_window_pixel_coords(self, raster, window, poly):
        # converts to window coords from utm
        local_poly = []
        for point in poly:
            local_poly.append(~raster.window_transform(window) * (point[0], point[1]))

        return local_poly

    def _load_geojsonfile(self, filename):
        with open(filename, 'r') as f:
            content = f.read()
            json = geojson.loads(content)
        return json

    def convert_poly(self, i, transformers):
        img_id = self.feat_to_img_id[i]
        transformer = transformers[img_id]

        geom = self.all_polygons[i]["geometry"]

        if geom["type"] == "MultiPolygon":
            for poly in geom["coordinates"]:
                for part in poly:
                    for point in part:
                        point[0], point[1] = list(transformer.transform(point[0], point[1]))
        elif geom["type"] == "Polygon":
            for poly in geom["coordinates"]:
                for point in poly:
                    point[0], point[1] = list(transformer.transform(point[0], point[1]))

    def augmentation(self, enable=True):
        self.augment = enable


class RasterizedOpenCities(BuildingDataset):

    def __init__(self,
                 tier=1,
                 show_mode=False,
                 augment=False,
                 small_subset=False,
                 crop_size=1024,
                 resize_size=224,
                 window_random_shift=2048,
                 data_dir="./",
                 baseline_mode=False,
                 transform=None,
                 val=False,
                 val_split=0.1,
                 split_seed=42,
                 sampling_mode="polygons"):

        super().__init__(tier, show_mode, augment, small_subset, crop_size,
                         resize_size, window_random_shift, data_dir, baseline_mode, transform, val, val_split, split_seed, sampling_mode)

        self.img_ids_to_label_raster = dict()
        with open(data_dir + "/train_metadata.csv", 'r') as f:
            csvreader = csv.reader(f)
            header = next(csvreader)  # NOQA
            for row in csvreader:
                if int(row[3]) <= tier:
                    imgid = row[0].split("/")[2]
                    self.img_ids_to_label_raster[imgid] = row[1][:-7] + "tif"

    def __getitem__(self, i):

        if self.sampling_mode == "random":
            window, raster_id = self._get_random_window()
            img_id = raster_id
            raster = rasterio.open(self.data_dir + "/" + self.img_ids_to_tif[raster_id])
            label_raster = rasterio.open(self.data_dir + "/" + self.img_ids_to_label_raster[raster_id])
        else:
            img_id = self.feat_to_img_id[i]
            raster = rasterio.open(self.data_dir + "/" + self.img_ids_to_tif[img_id])
            label_raster = rasterio.open(self.data_dir + "/" + self.img_ids_to_label_raster[img_id])
            window = self._get_window(raster, self.all_polygons[i])

        b1, b2, b3, b4 = raster.read(window=window)
        out_image = np.stack([b1, b2, b3])
        out_image = np.moveaxis(out_image, 0, 2)[:, :, 0:3]
        img = Image.fromarray(out_image, 'RGB')

        if self.augment:
            trans = self.aug_transforms
        else:
            trans = self.noaug_transforms

        fill, edges, vertices, angles = label_raster.read(window=window)
        masks = np.stack([fill, edges, vertices])

        mean = [self.stats[img_id]["stats"]["mean_r"], self.stats[img_id]["stats"]["mean_g"], self.stats[img_id]["stats"]["mean_b"]]
        std = [self.stats[img_id]["stats"]["std_r"], self.stats[img_id]["stats"]["std_g"], self.stats[img_id]["stats"]["std_b"]]

        masks = np.moveaxis(masks, 0, 2)
        masks = Image.fromarray(masks, "RGB")
        masks = masks.resize((self.resize_size, self.resize_size))
        angles = Image.fromarray(angles)
        angles = angles.resize((self.resize_size, self.resize_size))

        if self.show_mode:
            masks = np.moveaxis(np.array(masks), 0, 2)
            return {"image": trans(img), "gt_polygons_image": masks, "gt_crossfield_angle": angles, "image_mean": mean,
                    "image_std": std}
        elif self.baseline_mode:
            masks = np.array(masks, dtype=np.float) / 255
            return {"image": self.img_transforms(trans(img)), "gt_polygons_image": masks}
        else:
            out_img = np.array(trans(img))
            out_img = torch.Tensor(np.moveaxis(out_img, -1, 0))
            masks = np.moveaxis(np.array(masks), 2, 0)
            masks = torch.Tensor(np.array(masks))
            angles = np.expand_dims(angles, 0)
            angles = torch.Tensor(np.array(angles))
            return {"image": out_img,
                    "gt_polygons_image": masks,
                    "gt_crossfield_angle": angles,
                    "image_mean": torch.Tensor(mean),
                    "image_std": torch.Tensor(std),
                    "name": str(i),
                    "original_image": img_id}


class OpenCitiesTestDataset(Dataset):

    def __init__(self, img_dir, output_size):
        super().__init__()
        self.tif_files = glob.glob(img_dir + "/*/*.tif")
        print("Found", str(len(self.tif_files)), "test images")

        self.transforms = transforms.Compose([transforms.Resize((output_size, output_size))])

    def __len__(self):
        return len(self.tif_files)

    def __getitem__(self, i):
        img = self.get_img(i)
        imgscaled = img / 255.0
        return {"image": img,
                "image_filepath": self.tif_files[i],
                "name": self.get_id(i),
                "image_mean": torch.mean(imgscaled, (1, 2)),
                "image_std" : torch.std(imgscaled, (1, 2))}

    def get_id(self, i):

        path = self.tif_files[i]
        return path.split("/")[-2]

    def get_img(self, i):

        img = Image.open(self.tif_files[i])
        img = img.convert("RGB")
        img = np.moveaxis(np.array(self.transforms(img)), 2, 0)

        return torch.Tensor(img)


def main():
    ds = RasterizedOpenCities(show_mode=True, tier=1, small_subset=False, sampling_mode="random")

    n_samples = int(sys.argv[1]) if len(sys.argv) > 1 else 1

    for i in range(n_samples):
        break
        n = random.randint(0, len(ds))

        sample = ds[n]

        fig = plt.figure()

        fig.add_subplot(1, 3, 1)
        plt.imshow(sample["image"])
        fig.add_subplot(1, 3, 2)
        plt.imshow(sample["gt_polygons_image"])
        fig.add_subplot(1, 3, 3)
        plt.imshow(sample["gt_crossfield_angle"])
        plt.show()

    print("Testing whole dataset")
    for i in tqdm(range(len(ds))):
        sample = ds[i]


if __name__ == "__main__":
    main()



def main():
    import cmd_queue
    import ubelt as ub

    queue = cmd_queue.Queue.create(backend="tmux", size=4)
    queue.add_header_command(
        ub.codeblock(
            """
            pyenv shell 3.10.5
            source $(pyenv prefix)/envs/pyenv-geowatch/bin/activate
            """
        )
    )
    template = ub.codeblock(
        r"""
        python -m geowatch.tasks.poly_from_point.predict \
        --method 'sam' \
        --filepath_to_images "{file_path_to_images}" \
        --filepath_to_region "{file_path_to_region}" \
        """
    )
    region_model_path = ub.Path(
        "/mnt/ssd2/data/dvc-repos/smart_phase3_data/annotations/drop8/region_models/"
    )
    region_model_list = list(region_model_path.glob("*.geojson"))
    kwcoco_dpath = ub.Path(
        "/mnt/ssd2/data/dvc-repos/smart_phase3_data/Aligned-Drop8-ARA/"
    )
    for region_path in region_model_list:
        region_id = region_path.stem
        kwcoco_path = (
            kwcoco_dpath / region_id / (f"imgonly-{region_id}-rawbands.kwcoco.zip")
        )
        if kwcoco_path.exists():
            fmtkw = {
                "file_path_to_images": kwcoco_path,
                "file_path_to_region": region_path,
            }
            cmd = template.format(**fmtkw)
            queue.submit(cmd, name=f"From point: {region_id} ")
    queue.print_commands()
    queue.run()


if __name__ == "__main__":
    main()



import torch
import torch.nn as nn
from .unet_dropout import up, outconv, inconv, down_blur


class positional_encoding(nn.Module):
    def __init__(self, dimensions=64):
        super(positional_encoding, self).__init__()
        self.dimensions = dimensions

    def __call__(self, x):
        embedding = []
        for t in range(self.dimensions // 2):
            embedding.append(torch.sin(x / 2.**t))
            embedding.append(torch.cos(x / 2.**t))
        return torch.stack(embedding, dim=-1)


class attention_unet(nn.Module):
    def __init__(self, in_channels, out_channels, dropout_rate=.2, pos_encode=False, positional_layers=None, attention_layers=1, mode='addition'):
        super(attention_unet, self).__init__()

        self.mode = mode
        if isinstance(attention_layers, int):
            self.attention_layers = list(range(1, attention_layers + 1))
        else:
            self.attention_layers = list(attention_layers)

        num_attention_layers = len(self.attention_layers)

        if not pos_encode:
            self.positional_layers = []
        elif not positional_layers:
            self.positional_layers = self.attention_layers
        else:
            self.positional_layers = positional_layers

        assert (num_attention_layers <= 8)

        if self.mode == 'concatenation':
            self.pos_embed_dim = 16
            dimension_adjustment = self.pos_embed_dim
            # self.positional_encoding = positional_encoding(dimensions=self.pos_embed_dim)
        else:
            # self.positional_encoding = positional_encoding(dimensions=self.feature_dimensions[0])
            dimension_adjustment = 0

        self.feature_dimensions = [32, 64, 128, 256, 512, 256, 128, 64, 32]
        self.attention_dimensions = [32, 64, 128, 256, 256, 128, 64, 32]

        self.dimension_adjustment = [0, 0, 0, 0, 0, 0, 0, 0, 0]

        for x in self.positional_layers:
            self.dimension_adjustment[x - 1] += dimension_adjustment
            if x < 5:
                self.dimension_adjustment[8 - x] += dimension_adjustment
        self.pos_encode = pos_encode

        self.inc = inconv(in_channels, self.feature_dimensions[0])
        #x1
        self.down1 = down_blur(self.feature_dimensions[0] + self.dimension_adjustment[0], self.feature_dimensions[1])
        #x2
        self.down2 = down_blur(self.feature_dimensions[1] + self.dimension_adjustment[1], self.feature_dimensions[2])
        #x3
        self.down3 = down_blur(self.feature_dimensions[2] + self.dimension_adjustment[2], self.feature_dimensions[3])
        #x4
        self.down4 = down_blur(self.feature_dimensions[3] + self.dimension_adjustment[3], self.feature_dimensions[3])
        #x5
        self.up1 = up(self.feature_dimensions[4] + self.dimension_adjustment[4], self.feature_dimensions[2])
        #x6
        self.up2 = up(self.feature_dimensions[3] + self.dimension_adjustment[5], self.feature_dimensions[1])
        #x7
        self.up3 = up(self.feature_dimensions[2] + self.dimension_adjustment[6], self.feature_dimensions[0])
        #x8
        self.up4 = up(self.feature_dimensions[1] + self.dimension_adjustment[7], self.feature_dimensions[0])

        self.outc = outconv(self.feature_dimensions[0], out_channels)

        self.drop = nn.Dropout(p=dropout_rate)

        self.query = nn.ModuleList()
        self.key = nn.ModuleList()
        self.value = nn.ModuleList()

        self.positional_encoding = nn.ModuleList()
        for k in range(self.attention_layers[-1]):
            self.query.append(nn.Conv2d(self.attention_dimensions[k], self.attention_dimensions[k], kernel_size=1))
            self.key.append(nn.Conv2d(self.attention_dimensions[k], self.attention_dimensions[k], kernel_size=1))
            self.value.append(nn.Conv2d(self.attention_dimensions[k], self.attention_dimensions[k], kernel_size=1))
            if pos_encode:
                if self.mode == 'concatenation':
                    self.positional_encoding.append(positional_encoding(dimensions=self.pos_embed_dim))
                else:
                    self.positional_encoding.append(positional_encoding(dimensions=self.attention_dimensions[k]))

    def forward(self, images, timestamps=None):
        # images # B x T x C x H x W

        B, T, c, h, w = images.shape
        images = images.view(B * T, c, h, w)
        x1 = self.inc(images)

        ### Attention
        if 1 in self.attention_layers:
            #B x T x 64 x H x W
            _, _, H, W = x1.shape
            Q = self.query[0](x1).view(B, T, -1, H, W)
            K = self.key[0](x1).view(B, T, -1, H, W)
            V = self.value[0](x1).view(B, T, -1, H, W)
            att_linear = torch.einsum('bpchw,bqchw->bpqhw', Q, K)
            att = torch.softmax(att_linear, dim=1)
            x1 = torch.einsum('bpqhw, bpchw -> bqchw', att, V)
            x1 = x1.contiguous().view(B * T, -1, H, W)
            if 1 in self.positional_layers:
                if timestamps is None:
                    if self.mode == 'concatenation':
                        positional_encode = torch.zeros(B * T, -1).to(x1.device)
                    else:
                        positional_encode = torch.zeros(B * T, self.attention_dimensions[0]).to(x1.device)
                else:
                    positional_encode = self.positional_encoding[0](timestamps).view(B * T, -1)

                positional_encode = positional_encode.unsqueeze(-1).unsqueeze(-1).repeat(1, 1, h, w)

                if self.mode == 'concatenation':
                    x1 = torch.cat([x1, positional_encode], dim=1)
                else:
                    x1 = x1 + positional_encode

        ###
        x2 = self.drop(self.down1(x1))
        ### Attention
        if 2 in self.attention_layers:
            _, C, H, W = x2.shape
            Q = self.query[1](x2).view(B, T, -1, H, W)
            K = self.key[1](x2).view(B, T, -1, H, W)
            V = self.value[1](x2).view(B, T, -1, H, W)
            att_linear = torch.einsum('bpchw,bqchw->bpqhw', Q, K)
            att = torch.softmax(att_linear, dim=1)
            x2 = torch.einsum('bpqhw, bpchw -> bqchw', att, V)
            x2 = x2.contiguous().view(B * T, -1, H, W)

            if 2 in self.positional_layers:
                if timestamps is None:
                    if self.mode == 'concatenation':
                        positional_encode = torch.zeros(B * T, -1).to(x1.device)
                    else:
                        positional_encode = torch.zeros(B * T, self.attention_dimensions[1]).to(x1.device)
                else:
                    positional_encode = self.positional_encoding[1](timestamps).view(B * T, -1)

                positional_encode = positional_encode.unsqueeze(-1).unsqueeze(-1).repeat(1, 1, H, W)

                if self.mode == 'concatenation':
                    x2 = torch.cat([x2, positional_encode], dim=1)
                else:
                    x2 = x2 + positional_encode
            ###

        x3 = self.drop(self.down2(x2))

        ### Attention
        if 3 in self.attention_layers:
            _, C, H, W = x3.shape
            Q = self.query[2](x3).view(B, T, -1, H, W)
            K = self.key[2](x3).view(B, T, -1, H, W)
            V = self.value[2](x3).view(B, T, -1, H, W)
            att_linear = torch.einsum('bpchw,bqchw->bpqhw', Q, K)
            att = torch.softmax(att_linear, dim=1)
            x3 = torch.einsum('bpqhw, bpchw -> bqchw', att, V)
            x3 = x3.contiguous().view(B * T, -1, H, W)
            if 3 in self.positional_layers:
                if timestamps is None:
                    if self.mode == 'concatenation':
                        positional_encode = torch.zeros(B * T, -1).to(x1.device)
                    else:
                        positional_encode = torch.zeros(B * T, self.attention_dimensions[2]).to(x1.device)
                else:
                    positional_encode = self.positional_encoding[2](timestamps).view(B * T, -1)

                positional_encode = positional_encode.unsqueeze(-1).unsqueeze(-1).repeat(1, 1, H, W)

                if self.mode == 'concatenation':
                    x3 = torch.cat([x3, positional_encode], dim=1)
                else:
                    x3 = x3 + positional_encode
            ###

        x4 = self.drop(self.down3(x3))

        ### Attention
        if 4 in self.attention_layers:
            _, C, H, W = x4.shape
            Q = self.query[3](x4).view(B, T, -1, H, W)
            K = self.key[3](x4).view(B, T, -1, H, W)
            V = self.value[3](x4).view(B, T, -1, H, W)
            att_linear = torch.einsum('bpchw,bqchw->bpqhw', Q, K)
            att = torch.softmax(att_linear, dim=1)
            x4 = torch.einsum('bpqhw, bpchw -> bqchw', att, V)
            x4 = x4.contiguous().view(B * T, -1, H, W)
            if 4 in self.positional_layers:
                if timestamps is None:
                    if self.mode == 'concatenation':
                        positional_encode = torch.zeros(B * T, -1).to(x1.device)
                    else:
                        positional_encode = torch.zeros(B * T, self.attention_dimensions[3]).to(x1.device)
                else:
                    positional_encode = self.positional_encoding[3](timestamps).view(B * T, -1)

                positional_encode = positional_encode.unsqueeze(-1).unsqueeze(-1).repeat(1, 1, H, W)

                if self.mode == 'concatenation':
                    x4 = torch.cat([x4, positional_encode], dim=1)
                else:
                    x4 = x4 + positional_encode

        ###
        x5 = self.drop(self.down4(x4))

        ## 5
        if 5 in self.attention_layers:
            _, C, H, W = x5.shape
            Q = self.query[4](x5).view(B, T, -1, H, W)
            K = self.key[4](x5).view(B, T, -1, H, W)
            V = self.value[4](x5).view(B, T, -1, H, W)
            att_linear = torch.einsum('bpchw,bqchw->bpqhw', Q, K)
            att = torch.softmax(att_linear, dim=1)
            x5 = torch.einsum('bpqhw, bpchw -> bqchw', att, V)
            x5 = x5.contiguous().view(B * T, -1, H, W)
            if 5 in self.positional_layers:
                if timestamps is None:
                    if self.mode == 'concatenation':
                        positional_encode = torch.zeros(B * T, -1).to(x1.device)
                    else:
                        positional_encode = torch.zeros(B * T, self.attention_dimensions[4]).to(x1.device)
                else:
                    positional_encode = self.positional_encoding[4](timestamps).view(B * T, -1)

                positional_encode = positional_encode.unsqueeze(-1).unsqueeze(-1).repeat(1, 1, H, W)

                if self.mode == 'concatenation':
                    x5 = torch.cat([x5, positional_encode], dim=1)
                else:
                    x5 = x5 + positional_encode
            ###

        x6 = self.drop(self.up1(x5, x4))  # make sure the right residual connections connect

        ## 6
        if 6 in self.attention_layers:
            _, C, H, W = x6.shape
            Q = self.query[5](x6).view(B, T, -1, H, W)
            K = self.key[5](x6).view(B, T, -1, H, W)
            V = self.value[5](x6).view(B, T, -1, H, W)
            att_linear = torch.einsum('bpchw,bqchw->bpqhw', Q, K)
            att = torch.softmax(att_linear, dim=1)
            x6 = torch.einsum('bpqhw, bpchw -> bqchw', att, V)
            x6 = x6.contiguous().view(B * T, -1, H, W)
            if 6 in self.positional_layers:
                if timestamps is None:
                    if self.mode == 'concatenation':
                        positional_encode = torch.zeros(B * T, -1).to(x1.device)
                    else:
                        positional_encode = torch.zeros(B * T, self.attention_dimensions[5]).to(x1.device)
                else:
                    positional_encode = self.positional_encoding[5](timestamps).view(B * T, -1)

                positional_encode = positional_encode.unsqueeze(-1).unsqueeze(-1).repeat(1, 1, H, W)

                if self.mode == 'concatenation':
                    x6 = torch.cat([x6, positional_encode], dim=1)
                else:
                    x6 = x6 + positional_encode
            ###

        x7 = self.drop(self.up2(x6, x3))

        ## 7
        if 7 in self.attention_layers:
            _, C, H, W = x7.shape
            Q = self.query[6](x7).view(B, T, -1, H, W)
            K = self.key[6](x7).view(B, T, -1, H, W)
            V = self.value[6](x7).view(B, T, -1, H, W)
            att_linear = torch.einsum('bpchw,bqchw->bpqhw', Q, K)
            att = torch.softmax(att_linear, dim=1)
            x7 = torch.einsum('bpqhw, bpchw -> bqchw', att, V)
            x7 = x7.contiguous().view(B * T, -1, H, W)
            if 7 in self.positional_layers:
                if timestamps is None:
                    if self.mode == 'concatenation':
                        positional_encode = torch.zeros(B * T, -1).to(x1.device)
                    else:
                        positional_encode = torch.zeros(B * T, self.attention_dimensions[6]).to(x1.device)
                else:
                    positional_encode = self.positional_encoding[6](timestamps).view(B * T, -1)

                positional_encode = positional_encode.unsqueeze(-1).unsqueeze(-1).repeat(1, 1, H, W)

                if self.mode == 'concatenation':
                    x7 = torch.cat([x7, positional_encode], dim=1)
                else:
                    x7 = x7 + positional_encode
            ###

        x8 = self.drop(self.up3(x7, x2))

        ## 8
        if 8 in self.attention_layers:
            _, C, H, W = x8.shape
            Q = self.query[7](x8).view(B, T, -1, H, W)
            K = self.key[7](x8).view(B, T, -1, H, W)
            V = self.value[7](x8).view(B, T, -1, H, W)
            att_linear = torch.einsum('bpchw,bqchw->bpqhw', Q, K)
            att = torch.softmax(att_linear, dim=1)
            x8 = torch.einsum('bpqhw, bpchw -> bqchw', att, V)
            x8 = x8.contiguous().view(B * T, -1, H, W)
            if 8 in self.positional_layers:
                if timestamps is None:
                    if self.mode == 'concatenation':
                        positional_encode = torch.zeros(B * T, -1).to(x1.device)
                    else:
                        positional_encode = torch.zeros(B * T, self.attention_dimensions[7]).to(x1.device)
                else:
                    positional_encode = self.positional_encoding[7](timestamps).view(B * T, -1)

                positional_encode = positional_encode.unsqueeze(-1).unsqueeze(-1).repeat(1, 1, H, W)

                if self.mode == 'concatenation':
                    x8 = torch.cat([x8, positional_encode], dim=1)
                else:
                    x8 = x8 + positional_encode
            ###

        x = self.up4(x8, x1)

        x = self.outc(x)

        return x.view(B, T, -1, h, w)



"""
ProgIter progress callback for lighting

Modified from the TQDM Progress Class, with this license:
# Copyright The Lightning AI team.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
import math
import os
import sys
from typing import Any, Dict, Optional, Union

from pytorch_lightning.utilities.types import STEP_OUTPUT

# check if ipywidgets is installed before importing tqdm.auto
# to ensure it won't fail and a progress bar is displayed
from progiter import ProgIter

import pytorch_lightning as pl
from pytorch_lightning.callbacks.progress.progress_bar import ProgressBar
from pytorch_lightning.utilities.rank_zero import rank_zero_debug

_PAD_SIZE = 5

_DEBUG = 0


class _ProgIter(ProgIter):
    def __init__(self, *args: Any, **kwargs: Any) -> None:
        """Custom progiter progressbar in case we need to customize"""
        # this just to make the make docs happy, otherwise it pulls docs which has some issues...
        kwargs.pop('smoothing', 0)  # API compatibility, unhandled, could implement
        super().__init__(*args, **kwargs)

    def reset(self, total=None):
        """
        Resets the progress to the start optionally with a new length
        """
        self.total = total
        self._reset_internals()

    @property
    def disable(self):
        return not self.enabled

    @property
    def n(self):
        """
        Alias for `self._iter_idx`
        """
        return self._iter_idx

    @n.setter
    def n(self, value):
        """
        TQDM allows the user to set 'n' to control number of iterations
        """
        self._iter_idx = value


class ProgIterProgressBar(ProgressBar):
    r"""
    This is the default progress bar used by Lightning. It prints to ``stdout`` using the
    :mod:`progiter` package and shows up to four different bars:

        - **sanity check progress:** the progress during the sanity check run
        - **train progress:** shows the training progress. It will pause if validation starts and will resume
          when it ends, and also accounts for multiple validation runs during training when
          :paramref:`~pytorch_lightning.trainer.trainer.Trainer.val_check_interval` is used.
        - **validation progress:** only visible during validation;
          shows total progress over all validation datasets.
        - **test progress:** only active when testing; shows total progress over all test datasets.

    For infinite datasets, the progress bar never ends.

    If you want to customize the default ``progiter`` progress bars used by
    Lightning, you can override specific methods of the callback class and pass
    your custom implementation to the
    :class:`~pytorch_lightning.trainer.trainer.Trainer`.

    CommandLine:
        xdoctest -m geowatch.utils.lightning_ext.callbacks.progiter_progress ProgIterProgressBar

    Example:
        >>> from geowatch.utils.lightning_ext.callbacks.progiter_progress import *  # NOQA
        >>> from geowatch.utils.lightning_ext.demo import LightningToyNet2d
        >>> from geowatch.utils.lightning_ext.callbacks import StateLogger
        >>> import ubelt as ub
        >>> from geowatch.utils.lightning_ext.monkeypatches import disable_lightning_hardware_warnings
        >>> disable_lightning_hardware_warnings()
        >>> default_root_dir = ub.Path.appdir('geowatch/lightning_ext/test/progiter_progress')
        >>> default_root_dir.delete().ensuredir()
        >>> # Test starting a model without any existing checkpoints
        >>> self = ProgIterProgressBar(time_thresh=1.0)
        >>> trainer = pl.Trainer(default_root_dir=default_root_dir, callbacks=[
        >>>     self,
        >>>     StateLogger()
        >>> ], max_epochs=5, accelerator='cpu', devices=1)
        >>> #model = LightningToyNet2d(num_train=5_000, num_val=5_000)
        >>> model = LightningToyNet2d()
        >>> trainer.fit(model)

    Args:
        refresh_rate: Determines at which rate (in number of batches) the progress bars get updated.
            Set it to ``0`` to disable the display.
        process_position: Set this to a value greater than ``0`` to offset the progress bars by this many lines.
            This is useful when you have progress bars defined elsewhere and want to show all of them
            together. This corresponds to
            :paramref:`~pytorch_lightning.trainer.trainer.Trainer.process_position` in the
            :class:`~pytorch_lightning.trainer.trainer.Trainer`.
    """

    def __init__(self, refresh_rate: int = 1, process_position: int = 0, time_thresh: float = 1.0):
        super().__init__()
        self._refresh_rate = self._resolve_refresh_rate(refresh_rate)
        self._process_position = process_position
        self._enabled = True
        self._train_progress_bar: Optional[ProgIter] = None
        self._val_progress_bar: Optional[ProgIter] = None
        self._test_progress_bar: Optional[ProgIter] = None
        self._predict_progress_bar: Optional[ProgIter] = None
        self._time_thresh = time_thresh

    def __getstate__(self) -> Dict:
        # can't pickle the ProgIter objects
        return {k: v if not isinstance(v, ProgIter) else None for k, v in vars(self).items()}

    @property
    def train_progress_bar(self) -> ProgIter:
        if self._train_progress_bar is None:
            raise TypeError(f"The `{self.__class__.__name__}._train_progress_bar` reference has not been set yet.")
        return self._train_progress_bar

    @train_progress_bar.setter
    def train_progress_bar(self, bar: ProgIter) -> None:
        self._train_progress_bar = bar

    @property
    def val_progress_bar(self) -> ProgIter:
        if self._val_progress_bar is None:
            raise TypeError(f"The `{self.__class__.__name__}._val_progress_bar` reference has not been set yet.")
        return self._val_progress_bar

    @val_progress_bar.setter
    def val_progress_bar(self, bar: ProgIter) -> None:
        self._val_progress_bar = bar

    @property
    def test_progress_bar(self) -> ProgIter:
        if self._test_progress_bar is None:
            raise TypeError(f"The `{self.__class__.__name__}._test_progress_bar` reference has not been set yet.")
        return self._test_progress_bar

    @test_progress_bar.setter
    def test_progress_bar(self, bar: ProgIter) -> None:
        self._test_progress_bar = bar

    @property
    def predict_progress_bar(self) -> ProgIter:
        if self._predict_progress_bar is None:
            raise TypeError(f"The `{self.__class__.__name__}._predict_progress_bar` reference has not been set yet.")
        return self._predict_progress_bar

    @predict_progress_bar.setter
    def predict_progress_bar(self, bar: ProgIter) -> None:
        self._predict_progress_bar = bar

    @property
    def refresh_rate(self) -> int:
        return self._refresh_rate

    @property
    def process_position(self) -> int:
        return self._process_position

    @property
    def is_enabled(self) -> bool:
        return self._enabled and self.refresh_rate > 0

    @property
    def is_disabled(self) -> bool:
        return not self.is_enabled

    def disable(self) -> None:
        self._enabled = False

    def enable(self) -> None:
        self._enabled = True

    def init_sanity_progiter(self) -> _ProgIter:
        """Override this to customize the ProgIter bar for the validation sanity run."""
        return _ProgIter(
            desc=self.sanity_check_description,
            # position=(2 * self.process_position),
            disable=self.is_disabled,
            # leave=False,
            # dynamic_ncols=True,
            file=sys.stdout,
            # homogeneous=False,
            time_thresh=self._time_thresh,
        )

    def init_train_progiter(self) -> _ProgIter:
        """Override this to customize the ProgIter bar for training."""
        return _ProgIter(
            desc=self.train_description,
            # position=(2 * self.process_position),
            disable=self.is_disabled,
            # leave=True,
            # dynamic_ncols=True,
            file=sys.stdout,
            # smoothing=0,
            homogeneous=False,
            time_thresh=self._time_thresh,
        )

    def init_predict_progiter(self) -> _ProgIter:
        """Override this to customize the ProgIter bar for predicting."""
        return _ProgIter(
            desc=self.predict_description,
            # position=(2 * self.process_position),
            disable=self.is_disabled,
            # leave=True,
            # dynamic_ncols=True,
            file=sys.stdout,
            # smoothing=0,
            homogeneous=False,
            time_thresh=self._time_thresh,
        )

    def init_validation_progiter(self) -> _ProgIter:
        """Override this to customize the ProgIter bar for validation."""
        # The train progress bar doesn't exist in `trainer.validate()`
        # has_main_bar = self.trainer.state.fn != "validate"
        return _ProgIter(
            desc=self.validation_description,
            # position=(2 * self.process_position + has_main_bar),
            disable=self.is_disabled,
            # leave=not has_main_bar,
            # dynamic_ncols=True,
            file=sys.stdout,
            homogeneous=False,
            time_thresh=self._time_thresh,
        )

    def init_test_progiter(self) -> _ProgIter:
        """Override this to customize the ProgIter bar for testing."""
        return _ProgIter(
            desc="Testing",
            # position=(2 * self.process_position),
            disable=self.is_disabled,
            # leave=True,
            # dynamic_ncols=True,
            file=sys.stdout,
            homogeneous=False,
            time_thresh=self._time_thresh,
        )

    def on_sanity_check_start(self, *_: Any) -> None:
        if _DEBUG:
            print('\n on_sanity_check_start \n')
        self.val_progress_bar = self.init_sanity_progiter()
        self.train_progress_bar = _ProgIter(disable=True)  # dummy progress bar

    def on_sanity_check_end(self, *_: Any) -> None:
        if _DEBUG:
            print('\n on_sanity_check_end \n')
        self.train_progress_bar.close()
        self.val_progress_bar.close()

    def on_train_start(self, *_: Any) -> None:
        if _DEBUG:
            print('\n on_train_start \n')
        self.train_progress_bar = self.init_train_progiter()

    def on_train_epoch_start(self, trainer: "pl.Trainer", *_: Any) -> None:
        if _DEBUG:
            print('\n on_train_epoch_start \n')
        self.train_progress_bar.reset(convert_inf(self.total_train_batches))
        self.train_progress_bar.begin()
        self.train_progress_bar.initial = 0
        self.train_progress_bar.set_description(f"Epoch {trainer.current_epoch}")

    def on_train_batch_end(
        self, trainer: "pl.Trainer", pl_module: "pl.LightningModule", outputs: STEP_OUTPUT, batch: Any, batch_idx: int
    ) -> None:
        n = batch_idx + 1
        if self._should_update(n, self.train_progress_bar.total):
            _update_n(self.train_progress_bar, n)
            self.train_progress_bar.set_postfix(self.get_metrics(trainer, pl_module), refresh=False)

    def on_train_epoch_end(self, trainer: "pl.Trainer", pl_module: "pl.LightningModule") -> None:
        if _DEBUG:
            print('\n on_train_epoch_end \n')
        if not self.train_progress_bar.disable:
            self.train_progress_bar.set_postfix(self.get_metrics(trainer, pl_module))
            self.train_progress_bar.close()

    def on_train_end(self, *_: Any) -> None:
        # self.train_progress_bar.close()
        if _DEBUG:
            print('\n on_train_end \n')

    def on_validation_start(self, trainer: "pl.Trainer", pl_module: "pl.LightningModule") -> None:
        if _DEBUG:
            print('\n on_validation_start \n')
        # Hack: on_train_epoch_end happens after on_validation_epoch_end, so
        # we close the train progress bar here if there is a validation set.
        if not self.train_progress_bar.disable:
            self.train_progress_bar.set_postfix(self.get_metrics(trainer, pl_module))
            self.train_progress_bar.close()

        if not trainer.sanity_checking:
            self.val_progress_bar = self.init_validation_progiter()

    def on_validation_batch_start(
        self,
        trainer: "pl.Trainer",
        pl_module: "pl.LightningModule",
        batch: Any,
        batch_idx: int,
        dataloader_idx: int = 0,
    ) -> None:
        if not self.has_dataloader_changed(dataloader_idx):
            return

        self.val_progress_bar.reset(convert_inf(self.total_val_batches_current_dataloader))
        self.val_progress_bar.begin()
        self.val_progress_bar.initial = 0
        desc = self.sanity_check_description if trainer.sanity_checking else self.validation_description
        self.val_progress_bar.set_description(f"{desc} DataLoader {dataloader_idx}")

    def on_validation_batch_end(
        self,
        trainer: "pl.Trainer",
        pl_module: "pl.LightningModule",
        outputs: Optional[STEP_OUTPUT],
        batch: Any,
        batch_idx: int,
        dataloader_idx: int = 0,
    ) -> None:
        n = batch_idx + 1
        if self._should_update(n, self.val_progress_bar.total):
            _update_n(self.val_progress_bar, n)

    def on_validation_end(self, trainer: "pl.Trainer", pl_module: "pl.LightningModule") -> None:
        if _DEBUG:
            print('\n on_validation_end \n')
        if self._train_progress_bar is not None and trainer.state.fn == "fit":
            self.train_progress_bar.set_postfix(self.get_metrics(trainer, pl_module))
        self.val_progress_bar.close()
        self.reset_dataloader_idx_tracker()

    def on_test_start(self, trainer: "pl.Trainer", pl_module: "pl.LightningModule") -> None:
        self.test_progress_bar = self.init_test_progiter()

    def on_test_batch_start(
        self,
        trainer: "pl.Trainer",
        pl_module: "pl.LightningModule",
        batch: Any,
        batch_idx: int,
        dataloader_idx: int = 0,
    ) -> None:
        if not self.has_dataloader_changed(dataloader_idx):
            return

        self.test_progress_bar.reset(convert_inf(self.total_test_batches_current_dataloader))
        self.test_progress_bar.begin()
        self.test_progress_bar.initial = 0
        self.test_progress_bar.set_description(f"{self.test_description} DataLoader {dataloader_idx}")

    def on_test_batch_end(
        self,
        trainer: "pl.Trainer",
        pl_module: "pl.LightningModule",
        outputs: Optional[STEP_OUTPUT],
        batch: Any,
        batch_idx: int,
        dataloader_idx: int = 0,
    ) -> None:
        n = batch_idx + 1
        if self._should_update(n, self.test_progress_bar.total):
            _update_n(self.test_progress_bar, n)

    def on_test_end(self, trainer: "pl.Trainer", pl_module: "pl.LightningModule") -> None:
        self.test_progress_bar.close()
        self.reset_dataloader_idx_tracker()

    def on_predict_start(self, trainer: "pl.Trainer", pl_module: "pl.LightningModule") -> None:
        self.predict_progress_bar = self.init_predict_progiter()

    def on_predict_batch_start(
        self,
        trainer: "pl.Trainer",
        pl_module: "pl.LightningModule",
        batch: Any,
        batch_idx: int,
        dataloader_idx: int = 0,
    ) -> None:
        if not self.has_dataloader_changed(dataloader_idx):
            return

        self.predict_progress_bar.reset(convert_inf(self.total_predict_batches_current_dataloader))
        self.predict_progress_bar.begin()
        self.predict_progress_bar.initial = 0
        self.predict_progress_bar.set_description(f"{self.predict_description} DataLoader {dataloader_idx}")

    def on_predict_batch_end(
        self,
        trainer: "pl.Trainer",
        pl_module: "pl.LightningModule",
        outputs: Any,
        batch: Any,
        batch_idx: int,
        dataloader_idx: int = 0,
    ) -> None:
        n = batch_idx + 1
        if self._should_update(n, self.predict_progress_bar.total):
            _update_n(self.predict_progress_bar, n)

    def on_predict_end(self, trainer: "pl.Trainer", pl_module: "pl.LightningModule") -> None:
        self.predict_progress_bar.close()
        self.reset_dataloader_idx_tracker()

    def print(self, *args: Any, sep: str = " ", **kwargs: Any) -> None:
        active_progress_bar = None
        raise AssertionError

        if self._train_progress_bar is not None and not self.train_progress_bar.disable:
            active_progress_bar = self.train_progress_bar
        elif self._val_progress_bar is not None and not self.val_progress_bar.disable:
            active_progress_bar = self.val_progress_bar
        elif self._test_progress_bar is not None and not self.test_progress_bar.disable:
            active_progress_bar = self.test_progress_bar
        elif self._predict_progress_bar is not None and not self.predict_progress_bar.disable:
            active_progress_bar = self.predict_progress_bar

        if active_progress_bar is not None:
            s = sep.join(map(str, args))
            active_progress_bar.write(s, **kwargs)

    def _should_update(self, current: int, total: int) -> bool:
        # return self.is_enabled and (current % self.refresh_rate == 0 or current == total)
        return True

    @staticmethod
    def _resolve_refresh_rate(refresh_rate: int) -> int:
        if os.getenv("COLAB_GPU") and refresh_rate == 1:
            # smaller refresh rate on colab causes crashes, choose a higher value
            rank_zero_debug("Using a higher refresh rate on Colab. Setting it to `20`")
            return 20
        return refresh_rate


def convert_inf(x: Optional[Union[int, float]]) -> Optional[Union[int, float]]:
    """The tqdm doesn't support inf/nan values.

    We have to convert it to None.

    """
    if x is None or math.isinf(x) or math.isnan(x):
        return None
    return x


def _update_n(bar: ProgIter, value: int) -> None:
    if not bar.disable:
        bar.n = value
        bar._slow_path_step_body()
        # bar.refresh()
